#!/usr/bin/env python
# -*- coding: utf-8 -*-

r"""Python ♡ Nasy.

    |             *         *
    |                  .                .
    |           .                              登
    |     *                      ,
    |                   .                      至
    |
    |                               *          恖
    |          |\___/|
    |          )    -(             .           聖 ·
    |         =\ -   /=
    |           )===(       *
    |          /   - \
    |          |-    |
    |         /   -   \     0.|.0
    |  NASY___\__( (__/_____(\=/)__+1s____________
    |  ______|____) )______|______|______|______|_
    |  ___|______( (____|______|______|______|____
    |  ______|____\_|______|______|______|______|_
    |  ___|______|______|______|______|______|____
    |  ______|______|______|______|______|______|_
    |  ___|______|______|______|______|______|____

author   : Nasy https://nasy.moe
date     : Apr  7, 2023
email    : Nasy <nasyxx+python@gmail.com>
filename : basic.py
project  : jaxrie
license  : GPL-3.0+

Basic
"""

# Standard Library
from collections.abc import Callable
from dataclasses import dataclass, field

# Types
from jax.typing import ArrayLike

# JAX
import haiku as hk
import jax
import jax.numpy as jnp

# Local
from jaxrie.manifold import Manifold

Array = jax.Array


def init_wrap(
    manifold: Manifold, k: ArrayLike, init: hk.initializers.Initializer
) -> hk.initializers.Initializer:
  """Wrap the initializer."""
  return lambda shape, dtype: manifold.expmap0(init(shape, dtype), k)


@dataclass
class HAct(hk.Module):
  """Hyperbolic Activation Layer."""

  activation: Callable[[Array], Array]
  manifold: Manifold
  kin: ArrayLike
  kout: ArrayLike

  def __call__(self, x: Array) -> Array:
    """Apply the layer."""
    dtype = x.dtype
    m = self.manifold
    self.kin = jnp.array(self.kin, dtype=dtype)
    self.kout = jnp.array(self.kout, dtype=dtype)

    return m.proj(
        m.expmap0(self.activation(m.logmap0(x, self.kin)), self.kout),
        self.kout,
    )


@dataclass
class HLinear(hk.Module):
  """Hyperbolic Linear Layer."""

  out_features: int
  manifold: Manifold
  k: ArrayLike = -1.0
  train_k: bool = True
  with_bias: bool = True
  w_init: hk.initializers.Initializer | None = None
  b_init: hk.initializers.Initializer | None = None
  name: str | None = None

  def __call__(self, x: Array) -> Array:
    """Apply the layer."""
    dtype = x.dtype
    input_size = self.input_size = x.shape[-1]
    m = self.manifold

    self.k = jnp.array(self.k, dtype=dtype)
    if self.train_k:
      self.k = hk.get_parameter("rie_k", (), init=hk.initializers.Constant(self.k))

    w_init = self.w_init
    if w_init is None:
      stddev = 1.0 / jnp.sqrt(input_size)
      w_init = hk.initializers.TruncatedNormal(stddev=stddev)

    w = hk.get_parameter("rie_w", [input_size, self.out_features], dtype, init=w_init)

    out = m.proj(m.matmull(x, w, self.k), self.k)

    if self.with_bias:
      b_init = self.b_init or hk.initializers.Constant(0.0)
      b = hk.get_parameter("rie_b", [self.out_features], dtype, init=b_init)
      out = m.proj(m.adde(out, b, self.k), self.k)

    hk.set_state("rie_k", self.k)
    return out


@dataclass
class HGCN(hk.Module):
  """Hyperbolic Graph Convolutional Network."""

  out_features: int
  manifold: Manifold
  k: ArrayLike = -1.0
  train_k: bool = True
  with_bias: bool = True
  w_init: hk.initializers.Initializer | None = field(
      default=hk.initializers.VarianceScaling(1.0, "fan_avg", "uniform")
  )
  b_init: hk.initializers.Initializer | None = None
  name: str | None = None

  def __call__(self, x: Array, adj: Array) -> Array:
    """Apply the layer."""
    project = HLinear(
        self.out_features,
        self.manifold,
        self.k,
        self.train_k,
        self.with_bias,
        self.w_init,
        self.b_init,
        name="project",
    )

    out = project(x)
    self.k = project.k

    hk.set_state("rie_k", self.k)
    return self.manifold.proj(self.manifold.matmulr(adj, out, self.k), self.k)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

r"""Python ♡ Nasy.

    |             *         *
    |                  .                .
    |           .                              登
    |     *                      ,
    |                   .                      至
    |
    |                               *          恖
    |          |\___/|
    |          )    -(             .           聖 ·
    |         =\ -   /=
    |           )===(       *
    |          /   - \
    |          |-    |
    |         /   -   \     0.|.0
    |  NASY___\__( (__/_____(\=/)__+1s____________
    |  ______|____) )______|______|______|______|_
    |  ___|______( (____|______|______|______|____
    |  ______|____\_|______|______|______|______|_
    |  ___|______|______|______|______|______|____
    |  ______|______|______|______|______|______|_
    |  ___|______|______|______|______|______|____

author   : Nasy https://nasy.moe
date     : Apr 16, 2023
email    : Nasy <nasyxx+python@gmail.com>
filename : transform.py
project  : jaxrie
license  : GPL-3.0+

Transform of the data.


For haiku parameters like only.
  TODO: Find a better way to get the curvature.
"""
# Standard Library
from functools import partial

# Types
from typing import Any, NamedTuple, cast

# JAX
import haiku as hk
import jax
import jax.numpy as jnp
import optax
from optax import bias_correction, update_moment  # type: ignore[attr-defined]
from optax._src.utils import canonicalize_dtype, cast_tree

# Local
from .updates import apply_riemannian_updates, get_k
from jaxrie import Manifold

Array = jax.Array


def mix_opt(
    euc: optax.GradientTransformation, rie: optax.GradientTransformation
) -> optax.GradientTransformationExtraArgs:
  """Mix euclidean and riemannian optimizer."""
  euc_mask = partial(
      hk.data_structures.map, lambda _, name, _v: not name.startswith("rie_")
  )
  rie_mask = partial(
      hk.data_structures.map, lambda _, name, _v: name.startswith("rie_")
  )
  return optax.chain(
      optax.masked(euc, euc_mask),
      optax.masked(rie, rie_mask),
  )


def riemannian_scale(manifold: Manifold) -> optax.GradientTransformationExtraArgs:
  """Riemannian scale."""

  def init_fn(params: optax.Params) -> optax.OptState:
    del params
    return optax.ScaleState()

  def update_fn(
      grads: hk.Params,
      state: optax.OptState,
      params: hk.Params,
      states: hk.State,
      eps: float = 4e-3,
  ) -> tuple[hk.Params, optax.OptState]:
    """Update the scale."""
    return (
        jax.tree_util.tree_map_with_path(
            lambda ks, g, p: manifold.egrad2rgrad(p, g, get_k(states, ks), eps=eps),
            grads,
            params,
        ),
        state,
    )

  return optax.GradientTransformationExtraArgs(init_fn, update_fn)  # type: ignore[arg-type]  # noqa: E501


class RieAdamState(NamedTuple):
  """Riemannian Adam state."""

  count_: Array  # for mypy bug #3950: tuple has a method named count
  mu: optax.Updates
  nu: optax.Updates
  tau: optax.Updates


def scale_rie_by_adam(
    manifold: Manifold,
    b1: float = 0.9,
    b2: float = 0.999,
    eps: float = 1e-8,
    eps_root: float = 0.0,
    mu_dtype: Any | None = None,
) -> optax.GradientTransformationExtraArgs:
  """Riemannian scale by Adam."""
  mu_dtype = canonicalize_dtype(mu_dtype)

  def init_fn(params: optax.Params) -> RieAdamState:
    return RieAdamState(
        count_=jnp.zeros((), dtype=jnp.int32),
        mu=jax.tree_util.tree_map(lambda x: jnp.zeros_like(x, dtype=mu_dtype), params),
        nu=jax.tree_util.tree_map(lambda x: jnp.zeros_like(x, dtype=mu_dtype), params),
        tau=jax.tree_util.tree_map(lambda x: jnp.zeros_like(x, dtype=mu_dtype), params),
    )

  def update_fn(
      grads: hk.Params,
      state: RieAdamState,
      params: hk.Params,
      states: hk.State,
  ) -> tuple[hk.Params, RieAdamState]:
    """Update the scale."""
    count_inc = cast(Array, optax.safe_int32_increment(state.count_))

    mu = update_moment(grads, state.tau, b1, 1)

    sqnorm_grads = jax.tree_util.tree_map_with_path(
        lambda ks, g: manifold.sqnorm(g, get_k(states, ks)), grads
    )
    nu = update_moment(sqnorm_grads, state.nu, b2, 1)

    muhat = bias_correction(mu, b1, count_inc)
    nuhat = bias_correction(nu, b2, count_inc)

    updates = jax.tree_util.tree_map(
        lambda m, n: m / (jnp.sqrt(n + eps_root) + eps), muhat, nuhat
    )
    mu = cast_tree(mu, mu_dtype)

    new_params = apply_riemannian_updates(params, updates, states, manifold)

    tau = jax.tree_util.tree_map_with_path(
        lambda ks, p, np, m: manifold.ptransp(p, np, m, get_k(states, ks)),
        params,
        new_params,
        mu,
    )

    return updates, RieAdamState(count_inc, mu, nu, tau)

  return optax.GradientTransformationExtraArgs(init_fn, update_fn)  # type: ignore[arg-type]  # noqa: E501



#!/usr/bin/env python
# -*- coding: utf-8 -*-

r"""Python ♡ Nasy.

    |             *         *
    |                  .                .
    |           .                              登
    |     *                      ,
    |                   .                      至
    |
    |                               *          恖
    |          |\___/|
    |          )    -(             .           聖 ·
    |         =\ -   /=
    |           )===(       *
    |          /   - \
    |          |-    |
    |         /   -   \     0.|.0
    |  NASY___\__( (__/_____(\=/)__+1s____________
    |  ______|____) )______|______|______|______|_
    |  ___|______( (____|______|______|______|____
    |  ______|____\_|______|______|______|______|_
    |  ___|______|______|______|______|______|____
    |  ______|______|______|______|______|______|_
    |  ___|______|______|______|______|______|____

author   : Nasy https://nasy.moe
date     : Apr 16, 2023
email    : Nasy <nasyxx+python@gmail.com>
filename : alias.py
project  : jaxrie
license  : GPL-3.0+

Alias.
"""


# Types
from typing import Any

# JAX
import optax

# Local
from . import transform
from jaxrie import Manifold

ScalarOrSchedule = float | optax.Schedule


def _scale_by_learning_rate(
    learning_rate: ScalarOrSchedule, flip_sign: bool = True
) -> optax.GradientTransformation:
  """Update by learning rate."""
  m = -1 if flip_sign else 1
  if callable(learning_rate):
    return optax.scale_by_schedule(lambda count: m * learning_rate(count))  # type: ignore[operator]  # noqa: E501
  return optax.scale(m * learning_rate)


def riemannian_sgd(
    manifold: Manifold,
    learning_rate: ScalarOrSchedule,
) -> optax.GradientTransformationExtraArgs:
  """Run Riemannian Stochastic Gradient Descent optimizer."""
  return optax.chain(
      transform.riemannian_scale(manifold),
      _scale_by_learning_rate(learning_rate),
  )


def riemannian_adam(
    manifold: Manifold,
    learning_rate: ScalarOrSchedule,
    b1: float = 0.9,
    b2: float = 0.999,
    eps: float = 1e-8,
    eps_root: float = 1e-8,
    mu_dtype: Any | None = None,
) -> optax.GradientTransformationExtraArgs:
  """Run the Riemannian Adam optimizer."""
  return optax.chain(
      transform.riemannian_scale(manifold),
      transform.scale_rie_by_adam(
          manifold=manifold,
          b1=b1,
          b2=b2,
          eps=eps,
          eps_root=eps_root,
          mu_dtype=mu_dtype,
      ),
      _scale_by_learning_rate(learning_rate),
  )


def riemannian_adamw(
    manifold: Manifold,
    learning_rate: ScalarOrSchedule,
    b1: float = 0.9,
    b2: float = 0.999,
    eps: float = 1e-8,
    eps_root: float = 0.0,
    mu_dtype: Any | None = None,
    weight_decay: float = 1e-4,
    mask: Any | None = None,
) -> optax.GradientTransformationExtraArgs:
  """Run the Riemannian AdamW optimizer."""
  return optax.chain(
      transform.riemannian_scale(manifold),
      transform.scale_rie_by_adam(
          manifold=manifold,
          b1=b1,
          b2=b2,
          eps=eps,
          eps_root=eps_root,
          mu_dtype=mu_dtype,
      ),
      optax.add_decayed_weights(weight_decay, mask),
      _scale_by_learning_rate(learning_rate),
  )


def rsgd(
    manifold: Manifold,
    elr: ScalarOrSchedule,
    rlr: ScalarOrSchedule,
) -> optax.GradientTransformationExtraArgs:
  """Run Riemannian Stochastic Gradient Descent optimizer."""
  return transform.mix_opt(
      optax.sgd(elr),
      riemannian_sgd(manifold, rlr),
  )


def rie_adam(
    manifold: Manifold,
    elr: ScalarOrSchedule,
    rlr: ScalarOrSchedule,
    b1: float = 0.9,
    b2: float = 0.999,
) -> optax.GradientTransformationExtraArgs:
  """Run the Riemannian Adam optimizer."""
  return transform.mix_opt(
      optax.adam(elr, b1, b2),
      riemannian_adam(manifold, rlr, b1, b2),
  )


def rie_adamw(
    manifold: Manifold,
    elr: ScalarOrSchedule,
    rlr: ScalarOrSchedule,
    b1: float = 0.9,
    b2: float = 0.999,
    weight_decay: float = 1e-4,
    mask: Any | None = None,
) -> optax.GradientTransformationExtraArgs:
  """Run the Riemannian AdamW optimizer."""
  return transform.mix_opt(
      optax.adamw(elr, b1, b2, weight_decay=weight_decay, mask=mask),
      riemannian_adamw(manifold, rlr, b1, b2, weight_decay=weight_decay, mask=mask),
  )



# Local
from . import _math as math
from ._base import BaseManifold as Manifold
from ._euclidean import Euclidean
from ._stereographic import Stereographic

euclidean = Euclidean()
stereographic = Stereographic()


__all__ = [
    "Manifold",
    "Euclidean",
    "euclidean",
    "Stereographic",
    "stereographic",
    "math",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

r"""Python ♡ Nasy.

    |             *         *
    |                  .                .
    |           .                              登
    |     *                      ,
    |                   .                      至
    |
    |                               *          恖
    |          |\___/|
    |          )    -(             .           聖 ·
    |         =\ -   /=
    |           )===(       *
    |          /   - \
    |          |-    |
    |         /   -   \     0.|.0
    |  NASY___\__( (__/_____(\=/)__+1s____________
    |  ______|____) )______|______|______|______|_
    |  ___|______( (____|______|______|______|____
    |  ______|____\_|______|______|______|______|_
    |  ___|______|______|______|______|______|____
    |  ______|______|______|______|______|______|_
    |  ___|______|______|______|______|______|____

author   : Nasy https://nasy.moe
date     : Apr 16, 2023
email    : Nasy <nasyxx+python@gmail.com>
filename : updates.py
project  : jaxrie
license  : GPL-3.0+

Updates
"""
# JAX
import haiku as hk
import jax
import jax.numpy as jnp
import optax

# Local
from jaxrie import Manifold

Array = jax.Array
DK = jax.tree_util.DictKey


def get_k(
    states: hk.State,
    ks: tuple[DK, DK],
    idx: int = 0,
    key: str = "rie_k",
) -> Array:
  """Get the curvature."""
  return states[str(ks[idx].key)][key]


def apply_riemannian_updates(
    params: hk.Params, updates: optax.Updates, states: hk.State, manifold: Manifold
) -> hk.Params:
  """Apply the riemannian update to the corresponding parameters."""
  return jax.tree_util.tree_map_with_path(
      lambda ks, p, u: jnp.asarray(
          manifold.expmap(p, u, get_k(states, ks)).astype(
              jnp.asarray(p).dtype
          ).squeeze()
      ),
      params,
      updates,
  )


def apply_mix_updates(
    params: hk.Params, updates: optax.Updates, states: hk.State, manifold: Manifold
) -> hk.Params:
  """Apply the mix update to the corresponding parameters."""

  def update_fn(ks: tuple[DK, DK], param: Array, update: Array) -> Array:
    """Update the params by the updates."""
    dtype = jnp.asarray(param).dtype
    if str(ks[1].key).startswith("rie_"):
      k = get_k(states, ks)
      return manifold.proj(
          manifold.expmap(param, update, k),
          k,
      ).astype(dtype).squeeze()
    return jnp.asarray(param + update, dtype=dtype)

  return jax.tree_util.tree_map_with_path(
      update_fn,
      params,
      updates,
  )



# Local
from .alias import (rie_adam, rie_adamw, riemannian_adam, riemannian_adamw,
                    riemannian_sgd, rsgd)
from .transform import get_k, mix_opt, riemannian_scale, scale_rie_by_adam
from .updates import apply_mix_updates, apply_riemannian_updates

__all__ = [
    "apply_mix_updates",
    "apply_riemannian_updates",
    "get_k",
    "mix_opt",
    "riemannian_scale",
    "scale_rie_by_adam",
    "riemannian_sgd",
    "rsgd",
    "riemannian_adam",
    "riemannian_adamw",
    "rie_adam",
    "rie_adamw",
]
